#include <stdio.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_vector.h>


// Some constants:

gsl_rng * RNG;

static const double steps = 500.0;
//static const int NinfSets = 100; // usually, 100 for mainSchools.c vs 50000 for chainSchools.c
//static const int NparamSets = 50000; // Must be equal to NinfSets when running chainSchools.c

static const double ini_alpha = 0.01; // Initialization of parameters
static const double ini_beta = 0.5;
static const double ini_gamma = 0.5;
static const double ini_Gta = 5.0;
static const double ini_Gtb = 5.0;

static const double lolim_cum = 0.025;
static const double uplim_cum = 0.975;


// Define structures of interest 

typedef struct{
    int Nindiv;
    int *IDperson; /* ID of the individual */
    int *IDfamily; /* ID of the household */
    int *IDcluster; /* ID of the outbreak */
    int *symonset; /* date of symptom onset (-1 if no symptoms) */
    int *diagnose; /* date of diagnosis */
    int *neg1; /* negative tests' dates */
    int *neg2;
    int *neg3;
    int *neg4;
    int *neg5;
    int *neg6;
    int *neg7;
    int *neg8;
    int *neg9;
    int *vax1; /* vaccination dates */
    int *vax2;
    int *vax3;
    int *attend; /* Does it attend school? 1 yes, 0 no. */
    int **idxFam; /* Each row is an individual, in order. Each column is a day, from 0 (14/09/2020) to 412 (31/10/2021). */
    int **idxSch; /* Refers to the presence (1) or not (0) of the individual to the setting on a specific day */
    int **idxCom; 
    long double **xi; /* Susceptibility of each individual on each day */
    double **rho; /* Transmissibility of each individual on each day */
} Indiv;

typedef struct{
    int Nindiv;
    int *IDperson;
    int *IDfamily;
    int *vax1;
    int *vax2;
    int *vax3;
    int **idxCom;
    long double **xi;
} Negs;

typedef struct{
    int *date; // 14/09/2020 is day 0, 31/10/2021 is day 412.
    double *incidence; // Incidence of each day in the community.
} Out;

typedef struct{
    long double alpha; /* Transmissibility in the community */
    long double beta; /* Transmissibility in the household */
    long double gamma; /* Transmissibility in the school */
    long double Gta; /* Mean parameter for the generation time gamma distribution */
    long double Gtb; /* Variance parameter for the generation time gamma distribution */
    double sigma_alpha;
    double sigma_beta;
    double sigma_gamma;
    double sigma_Gta;
    double sigma_Gtb;
    double *genDensity; /* Vector with the generation time density evaluated from 0 to MAX_GENTIME */
} Param;

typedef struct{
    int Nclusters;
    int NDAYS; /* 413 */
    int DAY0; /* 0 */
    int DAYL; /* 412 */
    double XI1; /* Parameters for the susceptibility */
    double XI2;
    double XI3;
    double W;
    double RHO; /* Parameter for the transmissibility */ 
    int MAX_GENTIME; /* Maximum generation time allowed */
    int MINT; /* Initial time of our simulations = min_infection_time - MAX_GENTME */
    int MAXT; /* Last time of our simulations = last_diagnosis */
} Global;


typedef struct{
    long double *probInf; /* associated to the infection of individual j */
    int *infector; /* Index i of infector of j. -1 if community infection, -999 if j is non-family individual and does not contribute to the likelihood */
    int *placeInf; /* 0 community, 1 houshold, 2 school. -999 if j is non-family individual and does not contribute to the likelihood */
} Chain;


// Functions:

/* Reads global variables, the number of individuals, the sigmas for the mcmc parameters */
void ReadFiles1(char *address_glo, Indiv *indiv, Negs *negs, Global *global, Param *param);

/* Reads individuals data, community incidence, and infection times */
void ReadFiles2(char *address_ind, char *address_neg, char *address_idxFam, char *address_idxSch, char *address_idxCom, char *address_idxComNeg,
                char *address_out, char *address_inf, Indiv *indiv, Negs *negs, Global global, Out *out, int ***infecTimes, int NinfSets);

/* Assigns the daily susceptibility to all positive individuals, from day 0 to last day (412) */
void Xi(Indiv *indiv, Global global);

/* Assigns the daily susceptibility to all negative individuals, from day 0 to last day (412) */
void XiNeg(Negs *negs, Global global);

/* Assigns the daily transmissibility to all positive individuals, from day 0 to last day (412) */
void Rho(Indiv *indiv, Global global);

/* Computes the vector for the generation time density given the parameters */
void GenDensity(Param *param, Global global);

/* Computes the different lambda's (forces of infection) from the community and from each household/school individual */
long double **ForcesInfec_OHS(Indiv indiv, Global global, Out out, Param param, int **infecTimes, int person, int setOfInfecTimes);

/* Computes the integral (positive) of the exponential in the Q term of the likelihood, for the positive individuals */
long double Qp_OHS(Indiv indiv, Global global, Out out, Param param, int **infecTimes, int person, int setOfInfecTimes);

/* Computes the integral (positive) of the exponential in the Q term of the likelihood, for the negative individuals */
long double Qn_OH(Indiv indiv, Negs negs, Global global, Out out, Param param, int **infecTimes, int person, int setOfInfecTimes);

/* Computes the log-likelihood given the parameter set */
long double Likelihood(Indiv indiv, Negs negs, Global global, Out out, Param param, Chain *chain, int **infecTimes, int setOfInfecTimes);

/* Assigns an infector to each individual */
void ChainRec(Indiv indiv, Global global, Out out, Param param, Chain *chain, int **infecTimes, int setOfInfecTimes);